import math
import numpy as np

from gpddatabase.MarkdownDataObjectCorrelationMatrix import MarkdownDataObjectCorrelationMatrix as MarkdownFunctionalities

from gpddatabase.Exceptions import ExceptionNoField
from gpddatabase.Exceptions import ExceptionWrongType
from gpddatabase.Exceptions import ExceptionWrongLength
from gpddatabase.Exceptions import ExceptionWrongLabel
from gpddatabase.Exceptions import ExceptionWrongCorrelationMatrix

class DataObjectCorrelationMatrix(MarkdownFunctionalities):

	'''Class representing correlation matrix.'''

	def __init__(self, data):

		#get label
		self.label = data[0]

		if  self.label is None:
			raise ExceptionNoField('label')

		if not isinstance(self.label, str):
			raise ExceptionWrongType('label')

		if len(self.label) > 40:
			raise ExceptionWrongLength('label', 40)

		if not self.label.replace('_', '').isalnum():
			raise ExceptionWrongLabel(self.label)

		#get size
		matrix_size = math.sqrt(len(data[1:]))

		if not matrix_size.is_integer():
			raise ExceptionWrongCorrelationMatrix(self.label)

		matrix_size = int(matrix_size)

		#check elements
		for element in data[1:]:

			if (not isinstance(element, int)) and (not isinstance(element, float)):
				raise ExceptionWrongCorrelationMatrix(self.label)

			if (element < -1.) or (element > 1.):
				raise ExceptionWrongCorrelationMatrix(self.label)

		#create matrix
		self.correlation_matrix = np.matrix(np.array(data[1:]).reshape((matrix_size, matrix_size)))

		#check if symmetric
		if not np.allclose(self.correlation_matrix, self.correlation_matrix.T):
			raise ExceptionWrongCorrelationMatrix(self.label)

		#check if 1 on diagonal
		for element in np.nditer(self.correlation_matrix.diagonal()):
			if element != 1.:
				raise ExceptionWrongCorrelationMatrix(self.label)

	def get_label(self):

		'''Get label of this matrix.'''

		return self.label

	def get_correlation_matrix(self):

		'''Get correlation matrix.'''

		return self.correlation_matrix



from gpddatabase.GenericTypes import GenericTypes

from gpddatabase.Exceptions import ExceptionNoField
from gpddatabase.Exceptions import ExceptionRequiredDiffSizes
from gpddatabase.Exceptions import ExceptionUnknownType

import gpddatabase as db

class DataTypes(GenericTypes):

	'''Class stroring data types'''

	def __init__(self, paths):

		#run parent constructor
		super().__init__(paths)

		#collect
		self.required_name = {}
		self.required_type = {}
		self.optional_name = {}
		self.optional_type = {}
		self.optional_default = {}

		for field in self.data:

			try:
				field['required_name']
			except KeyError as err:
				raise ExceptionNoField('required_name') from err

			try:
				field['required_type']
			except KeyError as err:
				raise ExceptionNoField('required_type') from err

			if len(field['required_name']) != len(field['required_type']):
				raise ExceptionRequiredDiffSizes(field['required_name'], field['required_type'])

			for value in field['required_type']:
				db.ExclusiveDatabase().get_required_types().check_type(value)

			self.required_name.update({field['name']: field['required_name']})
			self.required_type.update({field['name']: field['required_type']})

			if 'optional_name' in field:

				try:
					field['optional_type']
				except KeyError as err:
					raise ExceptionNoField('optional_type') from err

				try:
					field['optional_default']
				except KeyError as err:
					raise ExceptionNoField('optional_default') from err

				if len(field['optional_name']) != len(field['optional_type']):
					raise ExceptionRequiredDiffSizes(field['optional_name'], field['optional_type'])

				if len(field['optional_name']) != len(field['optional_default']):
					raise ExceptionRequiredDiffSizes(field['optional_name'], field['optional_default'])

				for value in field['optional_type']:
					db.ExclusiveDatabase().get_required_types().check_type(value)

				for value in field['optional_default']:
					i = field['optional_default'].index(value)

					db.ExclusiveDatabase().get_required_types().check_value(
						field['optional_type'][i],
						value
					)

				self.optional_name.update({field['name']: field['optional_name']})
				self.optional_type.update({field['name']: field['optional_type']})
				self.optional_default.update({field['name']: field['optional_default']})

	def get_required_name(self, value):

		'''Get names of required fields in conditions for a given data type.'''

		try:
			return self.required_name[value]
		except KeyError as err:
			raise ExceptionUnknownType(value) from err

	def get_required_type(self, value):

		'''Get types of required fields in conditions for a given data type.'''

		try:
			return self.required_type[value]
		except KeyError as err:
			raise ExceptionUnknownType(value) from err

	def get_required_type_by_name(self, valueA, valueB):

		'''Get type of required field in conditions for a given data type and condition name.'''

		try:
			i = self.required_name[valueA].index(valueB)
			return self.required_type[valueA][i]
		except KeyError as err:
			raise ExceptionUnknownType(valueA) from err

	def get_optional_name(self, value):

		'''Get names of optional fields in conditions for a given data type.'''

		try:
			return self.optional_name[value]
		except KeyError as err:
			raise ExceptionUnknownType(value) from err

	def get_optional_type(self, value):

		'''Get types of optional fields in conditions for a given data type.'''

		try:
			return self.optional_type[value]
		except KeyError as err:
			raise ExceptionUnknownType(value) from err

	def get_optional_type_by_name(self, valueA, valueB):

		'''Get type of optional field in conditions for a given data type and condition name.'''

		try:
			i = self.optional_name[valueA].index(valueB)
			return self.optional_type[valueA][i]
		except KeyError as err:
			raise ExceptionUnknownType(valueA) from err

	def get_optional_default_by_name(self, valueA, valueB):

		'''Get default of optional field in conditions for a given data type and condition name.'''

		try:
			i = self.optional_name[valueA].index(valueB)
			return self.optional_default[valueA][i]
		except KeyError as err:
			raise ExceptionUnknownType(valueA) from err



from gpddatabase.MarkdownDataObjectData import MarkdownDataObjectData as MarkdownFunctionalities

from gpddatabase.Exceptions import ExceptionNoField
from gpddatabase.Exceptions import ExceptionWrongType
from gpddatabase.Exceptions import ExceptionWrongLength
from gpddatabase.Exceptions import ExceptionWrongLabel
from gpddatabase.Exceptions import ExceptionNotUnique
from gpddatabase.Exceptions import ExceptionUnknownLabel

from gpddatabase.DataObjectDataSet import DataObjectDataSet

class DataObjectData(MarkdownFunctionalities):

	'''Class representing data object.'''

	def __init__(self, data):

		#get
		self.data_sets = {}

		for i_element, element in enumerate(data):

			#only data groups
			if 'data_set' not in element:
				continue

			#get label
			if 'label' in element['data_set']:

				label = element['data_set']['label']

				if label is None:
					raise ExceptionNoField('label')

				if not isinstance(label, str):
					raise ExceptionWrongType('label')

				if len(label) > 40:
					raise ExceptionWrongLength('label', 40)

				if not label.replace('_', '').isalnum():
					raise ExceptionWrongLabel(label)

			else:
				label = 'data_set_' + str(i_element)

			#store
			if label in self.data_sets:
				raise ExceptionNotUnique(label)

			self.data_sets.update({label: DataObjectDataSet(element['data_set'], label)})

	def get_data_sets(self):

		'''Get all data groups.'''

		return self.data_sets

	def get_data_set_labels(self):

		'''Get data set labels.'''

		return list(self.data_sets.keys())

	def get_data_set(self, label):

		'''Get a given data set by label.'''

		if label not in self.data_sets:
			raise ExceptionUnknownLabel(label)

		return self.data_sets[label]



from gpddatabase.MarkdownDataObjectUncertaintySet import MarkdownDataObjectUncertaintySet as MarkdownFunctionalities

from gpddatabase.Exceptions import ExceptionValuesLEQ
from gpddatabase.Exceptions import ExceptionValuesGEQ

from gpddatabase.DataObjectUncertainty import DataObjectUncertainty

class DataObjectUncertaintySet(MarkdownFunctionalities):

	'''Class representing a set of uncertainties including correlation matrix.'''

	def __init__(self, data):

		#values
		self.uncertainties = []

		if isinstance(data[0], str):

			offset = 1
			self.correlation_matrix = data[0]

		else:

			offset = 0
			self.correlation_matrix = None

		for uncertainty in data[offset:]:
			self.uncertainties.append(DataObjectUncertainty(uncertainty))

	def get_uncertainties(self):

		'''Get uncertainties.'''

		return self.uncertainties

	def get_correlation_matrix(self):

		'''Get correlation matrix.'''

		return self.correlation_matrix

	def get_number_of_uncertainties(self):

		'''Get number of stored uncertainties.'''

		return len(self.uncertainties)

	def get_uncertainty(self, i):

		'''Get a single uncertainty.'''

		#check index
		if i < 0:
			raise ExceptionValuesLEQ(i, -1)

		if i >= len(self.uncertainties):
			raise ExceptionValuesGEQ(i, len(self.uncertainties))

		return self.uncertainties[i]



from pathlib import Path as path

from gpddatabase.Exceptions import ExceptionNoDirectory
from gpddatabase.Exceptions import ExceptionNotWritable

#since object members in this class are created dynamically, disable pylint error with:
#pylint: disable=no-member
class MarkdownExclusiveDatabase:

	''' Markdown functionalities of ExclusiveDatabase class.'''

	def convert_to_markdown(self, directory):

		''' Get markdown representation.'''

		#result
		output = ''

		output += '# Available datasets' + '\n'
		output += '\n'
		output += '| uuid | collaboration | reference | type | pseudo | observables | comment |' + '\n'
		output += '| ---- | ------------- | --------- | ---- | ------ | ----------- | ------- |' + '\n'

		#check if directory exist
		if not path(directory).is_dir():
			raise ExceptionNoDirectory(directory)

		#loop over all files
		for uuid in self.get_uuids():

			#print status
			print("info: working for: ", uuid)

			#get object
			dataObject = self.get_data_object(uuid)

			#print
			output += '| ' + '[' + uuid + '](' + 'file_' + uuid + '.markdown' + ')'
			output += ' | ' + dataObject.get_general_info().get_collaboration()
			output += ' | ' + dataObject.get_general_info().get_reference()
			output += ' | ' + dataObject.get_general_info().get_data_type()
			output += ' | ' + str(dataObject.get_general_info().get_pseudodata())

			output += ' |'

			observableNames = []

			for label in dataObject.get_data().get_data_set_labels():
				for observableName in dataObject.get_data().get_data_set(label).get_observables().get_names():
					if observableName not in observableNames:
						observableNames.append(observableName)

			for observableName in observableNames:
				output += ' ' + observableName

			output += ' | ' + str(dataObject.get_general_info().get_comment())

			output += ' | ' + '\n'

			#write
			try:

				with open(directory + '/file_' + uuid + '.markdown', 'w', encoding="utf-8") as file:
					print(dataObject.convert_to_markdown(), file=file)

			except FileNotFoundError as err:
				raise ExceptionNotWritable(directory) from err

		#write
		try:

			with open(directory + '/data_sets.markdown', 'w', encoding="utf-8") as file:
				print(output, file=file)

		except FileNotFoundError as err:
			raise ExceptionNotWritable(directory) from err



#since object members in this class are created dynamically, disable pylint error with:
#pylint: disable=no-member
class MarkdownDataObjectCorrelationMatrix:

	''' Markdown representation of MarkdownDataObjectCorrelationMatrix class.'''

	def convert_to_markdown(self):

		''' Get markdown representation.'''

		return self.label + str(self.correlation_matrix)



#since object members in this class are created dynamically, disable pylint error with:
#pylint: disable=no-member
class MarkdownDataObject:

	''' Markdown representation of DataObject class.'''

	def convert_to_markdown(self):

		''' Get markdown representation.'''

		output = ''

		output += '# File: ' + self.uuid.convert_to_markdown() + '\n'

		output += '\n' + self.general_info.convert_to_markdown()

		output += '\n' + self.data.convert_to_markdown()

		if len(self.correlation.get_correlation_matrices()) != 0:
			output += '\n' + self.correlation.convert_to_markdown()

		return output



import yaml
from munch import Munch as munch

from gpddatabase.Exceptions import ExceptionNoDataInFile
from gpddatabase.Exceptions import ExceptionNoField
from gpddatabase.Exceptions import ExceptionNotUnique
from gpddatabase.Exceptions import ExceptionUnknownType

class GenericTypes:

	'''Generic definition of type.'''

	def __init__(self, paths):

		#data
		self.data = []

		#loop over paths
		for path in paths.split(":"):

			#get raw data
			with open(path, 'r', encoding="utf-8") as f:
				rawData = yaml.safe_load(f)

			#get data
			data = munch.fromDict(rawData)

			#check if not empty
			if data is None:
				raise ExceptionNoDataInFile(path)

			#check for data field
			try:
				data['data']
			except KeyError as err:
				raise ExceptionNoField('data') from err

			#append
			for field in data['data']:
				self.data.append(field)

		#collect
		self.types = []

		for field in self.data:

			try:
				field['name']
			except KeyError as err:
				raise ExceptionNoField('name') from err

			if field['name'] in self.types:
				raise ExceptionNotUnique(field['name'])

			self.types.append(field['name'])

		#collect
		self.descriptions = {}

		for field in self.data:

			try:
				field['description']
			except KeyError as err:
				raise ExceptionNoField('description') from err

			self.descriptions.update({field['name']: field['description']})

	def check_type(self, value):

		'''Check if type exist. If not, raise exception.'''

		if value not in self.types:
			raise ExceptionUnknownType(value)

	def get_description(self, value):

		'''Get description of a given type.'''

		try:
			return self.descriptions[value]
		except KeyError as err:
			raise ExceptionUnknownType(value) from err



import os
from pathlib import Path as path
import yaml
from munch import Munch as munch

from gpddatabase.MarkdownExclusiveDatabase import MarkdownExclusiveDatabase as MarkdownFunctionalities

from gpddatabase.RequiredTypes import RequiredTypes
from gpddatabase.UnitGroupTypes import UnitGroupTypes
from gpddatabase.UnitTypes import UnitTypes
from gpddatabase.VariableTypes import VariableTypes
from gpddatabase.DataTypes import DataTypes
from gpddatabase.ParticleTypes import ParticleTypes
from gpddatabase.ObservableTypes import ObservableTypes

from gpddatabase.DataObject import DataObject
from gpddatabase.DataObjectUUID import DataObjectUUID

from gpddatabase.Exceptions import ExceptionNoDirectory
from gpddatabase.Exceptions import ExceptionNoField
from gpddatabase.Exceptions import ExceptionNotUniqueUUID
from gpddatabase.Exceptions import ExceptionUnknownUUID

class ExclusiveDatabase(MarkdownFunctionalities):

	'''Main class representing the database. It is designed to be the singleton.'''

	instance = None

	def __new__(cls):

		#check if instance set
		if cls.instance is None:

			#create
			cls.instance = super(ExclusiveDatabase, cls).__new__(cls)

			#get path
			module_path = os.path.dirname(os.path.abspath(__file__))

			#initialisation
			cls.path_unit_group_types = module_path + '/data/types/unit_group_types.yaml'
			cls.path_unit_types = module_path + '/data/types/unit_types.yaml'
			cls.path_variable_types = module_path + '/data/types/variable_types.yaml'
			cls.path_data_types = module_path + '/data/types/data_types.yaml'
			cls.path_observable_types = module_path + '/data/types/observable_types.yaml'

         	#use ':' to specify multiple paths
			cls.path_data = module_path + '/data/DVCS' + ':' + module_path + '/data/latticeQCD' + ':' + module_path + '/data/other'

			#define and load types
			cls.required_types = None
			cls.unit_group_types = None
			cls.unit_types = None
			cls.variable_types = None
			cls.data_types = None
			cls.particle_types = None
			cls.observable_types = None

			cls.read_types(cls)

			#read uuids
			cls.files = None

			cls.read_uuids(cls)

		#return
		return cls.instance

	def read_types(self):

		'''Load database.'''

		self.required_types = RequiredTypes()
		self.unit_group_types = UnitGroupTypes(self.path_unit_group_types)
		self.unit_types = UnitTypes(self.path_unit_types)
		self.variable_types = VariableTypes(self.path_variable_types)
		self.data_types = DataTypes(self.path_data_types)
		self.particle_types = ParticleTypes()
		self.observable_types = ObservableTypes(self.path_observable_types)

		print("loading types... done")

	def read_uuids(self):

		'''Get list of availible database files.'''

		#clear
		self.files = {}

		#loop over directories
		for directory in self.path_data.split(":"):

			#check if directory exist
			if not path(directory).is_dir():
				raise ExceptionNoDirectory(directory)

			#get list of yaml files (recursively)
			files = path(directory).glob('**/*.yaml')

			#loop over files
			for file in files:

				#read
				with open(str(file), 'r', encoding="utf-8") as f:
					rawData = yaml.safe_load(f)

				data = munch.fromDict(rawData)

				#check
				if 'uuid' not in data:
					raise ExceptionNoField('uuid')

				uuid = DataObjectUUID(data['uuid']).get_uuid()

				#check if exist
				if uuid in self.files:
					raise ExceptionNotUniqueUUID(str(file))

				#save
				self.files.update({uuid: str(file)})

		print("loading uuids... done")

	def get_uuids(self):

		'''Get list of availible uuids.'''

		return list(self.files.keys())

	def get_data_object(self, uuid):

		'''Get DataObject corresponding to given UUID'''

		#look for path
		if uuid not in self.files:
			raise ExceptionUnknownUUID(uuid)

		#get path
		path_to_file = self.files[uuid]

		#read
		with open(path_to_file, 'r', encoding="utf-8") as f:
			rawData = yaml.safe_load(f)

		data = munch.fromDict(rawData)

		return DataObject(data)

	def get_path_to_unit_group_types(self):

		'''Get path to yaml file defining unit group types.'''

		return self.path_unit_group_types

	def get_path_to_unit_types(self):

		'''Get path to yaml file defining unit types.'''

		return self.path_unit_types

	def get_path_to_variable_types(self):

		'''Get path to yaml file defining variable types.'''

		return self.path_variable_types

	def get_path_to_data_types(self):

		'''Get path to yaml file defining data types.'''

		return self.path_data_types

	def get_path_to_observable_types(self):

		'''Get path to yaml file defining observable types.'''

		return self.path_observable_types

	def get_path_to_databse(self):

		'''Get path to database files.'''

		return self.path_data

	def set_path_to_unit_group_types(self, path_to_file):

		'''Set path to yaml file defining unit group types (triggers reloading database).'''

		self.path_unit_group_types = path_to_file
		self.read_types()

	def set_path_to_unit_types(self, path_to_file):

		'''Set path to yaml file defining unit types (triggers reloading database).'''

		self.path_unit_types = path_to_file
		self.read_types()

	def set_path_to_variable_types(self, path_to_file):

		'''Set path to yaml file defining variable types (triggers reloading database).'''

		self.path_variable_types = path_to_file
		self.read_types()

	def set_path_to_data_types(self, path_to_file):

		'''Set path to yaml file defining data types (triggers reloading database).'''

		self.path_data_types = path_to_file
		self.read_types()

	def set_path_to_observable_types(self, path_to_file):

		'''Set path to yaml file defining observable types (triggers reloading database).'''

		self.path_observable_types = path_to_file
		self.read_types()

	def set_path_to_databse(self, path_to_file):

		'''Set specific path to database files. After setting the path the list of availible files is automatically updated.'''

		self.path_data = path_to_file
		self.read_uuids()

	def get_required_types(self):

		'''Get object defining types of required attributes.'''

		return self.required_types

	def get_unit_group_types(self):

		'''Get object defining unit group types.'''

		return self.unit_group_types

	def get_unit_types(self):

		'''Get object defining unit types.'''

		return self.unit_types

	def get_variable_types(self):

		'''Get object defining variable types.'''

		return self.variable_types

	def get_data_types(self):

		'''Get object defining data types.'''

		return self.data_types

	def get_particle_types(self):

		'''Get object defining particle types.'''

		return self.particle_types

	def get_observable_types(self):

		'''Get object defining observable types.'''

		return self.observable_types



'''
Python macro for testing yaml files with given schema defined in json language.

usage: python3 test_schema.py file_to_be_tested.yaml schema.json
'''

import sys
import json
import yaml
from jsonschema import validate, ValidationError, FormatChecker

import datetime

def convert_dates(obj):
    if isinstance(obj, dict):
        return {k: convert_dates(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [convert_dates(item) for item in obj]
    elif isinstance(obj, datetime.date):
        return obj.isoformat()
    return obj

with open(sys.argv[1]) as stream:
    try:
        data = yaml.safe_load(stream)
    except yaml.YAMLError as exc:
        print(exc)

with open(sys.argv[2]) as stream:
    schema = json.load(stream)

try:  
    validate(instance=convert_dates(data), schema=schema, format_checker=FormatChecker())
except ValidationError as e:
    print(e.message)



from gpddatabase.MarkdownDataObjectKinematics import MarkdownDataObjectKinematics as MarkdownFunctionalities

from gpddatabase.Exceptions import ExceptionNoField
from gpddatabase.Exceptions import ExceptionNotUnique
from gpddatabase.Exceptions import ExceptionRequiredDiffSizes
from gpddatabase.Exceptions import ExceptionDifferentUnitGroups
from gpddatabase.Exceptions import ExceptionWrongType

import gpddatabase as db

from gpddatabase.DataObjectUncertaintySet import DataObjectUncertaintySet
from gpddatabase.DataObjectKinematicBin import DataObjectKinematicBin
from gpddatabase.DataObjectReplicaSet import DataObjectReplicaSet

class DataObjectKinematics(MarkdownFunctionalities):

	'''Class representing a data kinematics.'''

	def __init__(self, data):

		#get names of variables
		self.names = []

		if 'name' not in data:
			raise ExceptionNoField('name')

		if data['name'] is None:
			raise ExceptionNoField('name')

		for name in data['name']:

			db.ExclusiveDatabase().get_variable_types().check_type(name)

			if name in self.names:
				raise ExceptionNotUnique(name)

			self.names.append(name)

		#get units
		self.units = []

		if 'unit' not in data:
			raise ExceptionNoField('unit')

		if data['unit'] is None:
			raise ExceptionNoField('unit')

		if len(data['name']) != len(data['unit']):
			raise ExceptionRequiredDiffSizes(data['name'], data['unit'])

		for unit in data['unit']:
			db.ExclusiveDatabase().get_unit_types().check_type(unit)

		for name, unit in zip(data['name'], data['unit']):

			if db.ExclusiveDatabase().get_variable_types().get_unit_group(name) != db.ExclusiveDatabase().get_unit_types().get_unit_group(unit):
				raise ExceptionDifferentUnitGroups(name, unit)

			self.units.append(unit)

		#get values
		self.values = []

		if 'value' not in data:
			raise ExceptionNoField('value')

		if data['value'] is None:
			raise ExceptionNoField('value')

		for point in data['value']:

			if len(point) != len(self.names):
				raise ExceptionRequiredDiffSizes(point, self.names)

			for value in point:
				if (not isinstance(value, int)) and (not isinstance(value, float)):
					raise ExceptionWrongType('value')

			self.values.append(point)

		#get unc
		self.uncertainties = []

		if 'unc' in data:

			if data['unc'] is None:
				raise ExceptionNoField('unc')

			for uncertainty_set_raw in data['unc']:

				uncertainty_set = DataObjectUncertaintySet(uncertainty_set_raw)

				if len(uncertainty_set.get_uncertainties()) != len(self.names):
					raise ExceptionRequiredDiffSizes(uncertainty_set.get_uncertainties(), self.names)

				self.uncertainties.append(uncertainty_set)

			if len(self.uncertainties) != len(self.values):
				raise ExceptionRequiredDiffSizes(self.uncertainties, self.values)

		#get bins
		self.bins = []

		if 'bin' in data:

			if data['bin'] is None:
				raise ExceptionNoField('bin')

			for bins in data['bin']:

				this_bins = []

				if len(bins) != len(self.names):
					raise ExceptionRequiredDiffSizes(bins, self.names)

				for this_bin in bins:
					this_bins.append(DataObjectKinematicBin(this_bin))

				self.bins.append(this_bins)

			if len(self.bins) != len(self.values):
				raise ExceptionRequiredDiffSizes(self.bins, self.values)

		#get replicas
		self.replicas = []

		if 'replica' in data:

			if data['replica'] is None:
				raise ExceptionNoField('replica')

			for replicas in data['replica']:

				this_replicas = []

				if len(replicas) != len(self.names):
					raise ExceptionRequiredDiffSizes(replicas, self.names)

				for replica in replicas:
					this_replicas.append(DataObjectReplicaSet(replica))

				self.replicas.append(this_replicas)

			if len(self.replicas) != len(self.values):
				raise ExceptionRequiredDiffSizes(self.replicas, self.values)


	def get_names(self):

		'''Get names of variables.'''

		return self.names

	def get_units(self):

		'''Get names and units.'''

		return self.units

	def get_values(self):

		'''Get values.'''

		return self.values

	def get_uncertainties(self):

		'''Get uncertainties.'''

		return self.uncertainties

	def get_bins(self):

		'''Get bins.'''

		return self.bins

	def get_replicas(self):

		'''Get replicas.'''

		return self.replicas



#since object members in this class are created dynamically, disable pylint error with:
#pylint: disable=no-member
class MarkdownDataObjectData:

	''' Markdown representation of DataObjectData class.'''

	def convert_to_markdown(self):

		''' Get markdown representation.'''

		output = ''

		output += '## Data sets:' + '\n'

		output += '\n'

		for data_set in self.data_sets:
			output += self.data_sets[data_set].convert_to_markdown()

		return output



from gpddatabase.MarkdownDataObjectKinematicBin import MarkdownDataObjectKinematicBin as MarkdownFunctionalities

from gpddatabase.Exceptions import ExceptionNoField
from gpddatabase.Exceptions import ExceptionWrongLength
from gpddatabase.Exceptions import ExceptionWrongType
from gpddatabase.Exceptions import ExceptionValuesGEQ

class DataObjectKinematicBin(MarkdownFunctionalities):

	'''Class representing a single kinematic bin.'''

	def __init__(self, data):

		#min max
		if data is None:
			raise ExceptionNoField('bin')

		if len(data) != 2:
			raise ExceptionWrongLength('bin', 2)

		for value in data:

			if (not isinstance(value, int)) and (not isinstance(value, float)):
				raise ExceptionWrongType('bin')

		if data[0] >= data[1]:
			raise ExceptionValuesGEQ(data[0], data[1])

		self.min = data[0]
		self.max = data[1]

	def get_min(self):

		'''Get lower value.'''

		return self.min

	def get_max(self):

		'''Get higher value.'''

		return self.max



from gpddatabase.GenericTypes import GenericTypes

from gpddatabase.Exceptions import ExceptionNoField
from gpddatabase.Exceptions import ExceptionUnknownType

import gpddatabase as db

class VariableTypes(GenericTypes):

	'''Class stroring variable types'''

	def __init__(self, paths):

		#run parent constructor
		super().__init__(paths)

		#collect
		self.unit_groups = {}

		for field in self.data:

			try:
				field['unit_group']
			except KeyError as err:
				raise ExceptionNoField('unit_group') from err

			db.ExclusiveDatabase().get_unit_group_types().check_type(field['unit_group'])

			self.unit_groups.update({field['name']: field['unit_group']})

	def get_unit_group(self, value):

		'''Get unit group of a given variable.'''

		try:
			return self.unit_groups[value]
		except KeyError as err:
			raise ExceptionUnknownType(value) from err



from gpddatabase.MarkdownDataObject import MarkdownDataObject as MarkdownFunctionalities

from gpddatabase.Exceptions import ExceptionNoField
from gpddatabase.Exceptions import ExceptionUnknownLabel

from gpddatabase.DataObjectUUID import DataObjectUUID
from gpddatabase.DataObjectGeneralInfo import DataObjectGeneralInfo
from gpddatabase.DataObjectData import DataObjectData
from gpddatabase.DataObjectCorrelation import DataObjectCorrelation

class DataObject(MarkdownFunctionalities):

	'''Class representing a single data file.'''

	def __init__(self, data):

		#read uuid
		if 'uuid' not in data:
			raise ExceptionNoField('uuid')

		self.uuid = DataObjectUUID(data['uuid'])

		#read general info
		if 'general_info' not in data:
			raise ExceptionNoField('general_info')

		self.general_info = DataObjectGeneralInfo(data['general_info'])

		#read correlation
		if 'correlation' not in data:
			self.correlation = DataObjectCorrelation({})
		else:
			self.correlation = DataObjectCorrelation(data['correlation'])

		#read data
		if 'data' not in data:
			raise ExceptionNoField('data')

		self.data = DataObjectData(data['data'])

		#check labels of correlation matrices
		for data_set in self.data.get_data_sets():

			if self.data.get_data_sets()[data_set].get_kinematics().get_uncertainties():
				for uncertainty in self.data.get_data_sets()[data_set].get_kinematics().get_uncertainties():
					label = uncertainty.get_correlation_matrix()
					if label is not None:
						if label not in self.correlation.get_correlation_matrices():
							raise ExceptionUnknownLabel(label)

			if self.data.get_data_sets()[data_set].get_observables().get_stat_uncertainties():
				for uncertainty in self.data.get_data_sets()[data_set].get_observables().get_stat_uncertainties():
					label = uncertainty.get_correlation_matrix()
					if label is not None:
						if label not in self.correlation.get_correlation_matrices():
							raise ExceptionUnknownLabel(label)

			if self.data.get_data_sets()[data_set].get_observables().get_sys_uncertainties():
				for uncertainty in self.data.get_data_sets()[data_set].get_observables().get_sys_uncertainties():
					label = uncertainty.get_correlation_matrix()
					if label is not None:
						if label not in self.correlation.get_correlation_matrices():
							raise ExceptionUnknownLabel(label)

			if self.data.get_data_sets()[data_set].get_observables().get_sys_uncertainties_contribution():
				for uncertainties in self.data.get_data_sets()[data_set].get_observables().get_sys_uncertainties_contribution():
					for uncertainty in uncertainties:
						label = uncertainty.get_correlation_matrix()
						if label is not None:
							if label not in self.correlation.get_correlation_matrices():
								raise ExceptionUnknownLabel(label)

			if self.data.get_data_sets()[data_set].get_observables().get_norm_uncertainties():
				label = self.data.get_data_sets()[data_set].get_observables().get_norm_uncertainties().get_correlation_matrix()
				if label is not None:
					if label not in self.correlation.get_correlation_matrices():
						raise ExceptionUnknownLabel(label)

			if self.data.get_data_sets()[data_set].get_observables().get_norm_uncertainties_contribution():
				for uncertainty in self.data.get_data_sets()[data_set].get_observables().get_norm_uncertainties_contribution():
					label = uncertainty.get_correlation_matrix()
					if label is not None:
						if label not in self.correlation.get_correlation_matrices():
							raise ExceptionUnknownLabel(label)

	def get_uuid(self):

		'''Get UUID data object.'''

		return self.uuid

	def get_general_info(self):

		'''Get general information data object.'''

		return self.general_info

	def get_correlation(self):

		'''Get general information data object.'''

		return self.correlation

	def get_data(self):

		'''Get data object.'''

		return self.data



from gpddatabase.Exceptions import ExceptionUnknownType
from gpddatabase.Exceptions import ExceptionWrongType

import gpddatabase as db

class RequiredTypes:

	'''Class defining types of required fields and conversion function. '''

	def __init__(self):

		#define availible values
		self.required_types = {'bool', 'integer', 'float', 'string', 'particle', 'unit'}

	def check_type(self, value):

		'''Check if type exist. If not, raise exception.'''

		if value not in self.required_types:
			raise ExceptionUnknownType(value)

	def check_value(self, valueA, valueB):

		'''Check if given value corresponds to given type.'''

		if valueA == 'bool':
			if not isinstance(valueB, bool):
				raise ExceptionWrongType(valueB)
		elif valueA == 'integer':
			if not isinstance(valueB, int):
				raise ExceptionWrongType(valueB)
		elif valueA == 'float':
			if not isinstance(valueB, float):
				raise ExceptionWrongType(valueB)
		elif valueA == 'string':
			if not isinstance(valueB, str):
				raise ExceptionWrongType(valueB)
		elif valueA == 'particle':
			db.ExclusiveDatabase().get_particle_types().check_type(valueB)
		elif valueA == 'unit':
			db.ExclusiveDatabase().get_unit_types().check_type(valueB)



#since object members in this class are created dynamically, disable pylint error with:
#pylint: disable=no-member
class MarkdownDataObjectGeneralInfo:

	''' Markdown representation of DataObjectGeneralInfo class.'''

	def convert_to_markdown(self):

		''' Get markdown representation.'''

		output = ''

		output += '## General information:' + '\n'
		output += '\n'
		output += '* date: ' + str(self.date) + '\n'
		output += '* data type: ' + self.data_type + '\n'
		output += '* pseudodata: ' + ('true' if self.pseudodata else 'false') + '\n'
		output += '* collaboration: ' + self.collaboration + '\n'
		if self.reference:
			output += '* reference: ' + self.reference + '\n'

		output += '* conditions: ' + '\n'
		for condition in self.conditions:
			output += '   * ' + str(condition).replace("_", " ") + ': ' + str(self.conditions[condition]) + '\n'

		if self.comment:
			output += '* comment: ' + self.comment + '\n'

		return output



from gpddatabase.MarkdownDataObjectDataPoint import MarkdownDataObjectDataPoint as MarkdownFunctionalities

class DataObjectDataPoint(MarkdownFunctionalities):

	'''Class representing a single data point.'''

	def __init__(self,
		data_set_label,

		kinematic_names,
		kinematic_units,
		kinematic_values,
		kinematic_uncertainties,
		kinematic_bins,
		kinematic_replicas,

		observables_names,
		observables_units,
		observables_values,
		observables_stat_uncertainties,
		observables_sys_uncertainties,
		observables_sys_uncertainties_contribution_labels,
		observables_sys_uncertainties_contribution,
		observables_norm_uncertainties,
		observables_norm_uncertainties_contribution_labels,
		observables_norm_uncertainties_contribution,
		observables_replicas
		):

		self.data_set_label = data_set_label

		self.kinematic_names = kinematic_names
		self.kinematic_units = kinematic_units
		self.kinematic_values = kinematic_values
		self.kinematic_uncertainties = kinematic_uncertainties
		self.kinematic_bins = kinematic_bins
		self.kinematic_replicas = kinematic_replicas

		self.observables_names = observables_names
		self.observables_units = observables_units
		self.observables_values = observables_values
		self.observables_stat_uncertainties = observables_stat_uncertainties
		self.observables_sys_uncertainties = observables_sys_uncertainties
		self.observables_sys_uncertainties_contribution_labels = observables_sys_uncertainties_contribution_labels
		self.observables_sys_uncertainties_contribution = observables_sys_uncertainties_contribution
		self.observables_norm_uncertainties = observables_norm_uncertainties
		self.observables_norm_uncertainties_contribution_labels = observables_norm_uncertainties_contribution_labels
		self.observables_norm_uncertainties_contribution = observables_norm_uncertainties_contribution
		self.observables_replicas = observables_replicas

	def get_observables_names(self):

		'''Get names of observables.'''

		return self.observables_names

	def get_observables_units(self):

		'''Get units of observables.'''

		return self.observables_units

	def get_observables_values(self):

		'''Get values of observables.'''

		return self.observables_values

	def get_observables_stat_uncertainties(self):

		'''Get statistical uncertainties of observables.'''

		return self.observables_stat_uncertainties

	def get_observables_sys_uncertainties(self):

		'''Get systematic uncertainties of observables.'''

		return self.observables_sys_uncertainties

	def get_observables_sys_uncertainties_contribution_labels(self):

		'''Get labels of contributions to systematic uncertainties of observables.'''

		return self.observables_sys_uncertainties_contribution_labels

	def get_observables_sys_uncertainties_contribution(self):

		'''Get contribution to systematic uncertainties of observables.'''

		return self.observables_sys_uncertainties_contribution

	def get_observables_norm_uncertainties(self):

		'''Get normalisation uncertainties of observables.'''

		return self.observables_norm_uncertainties

	def get_observables_norm_uncertainties_contribution_labels(self):

		'''Get labels of contributions to normalisation uncertainties of observables.'''

		return self.observables_norm_uncertainties_contribution_labels

	def get_observables_norm_uncertainties_contribution(self):

		'''Get contribution to normalisation uncertainties of observables.'''

		return self.observables_norm_uncertainties_contribution

	def get_observables_replicas(self):

		'''Get replicas of observables.'''

		return self.observables_replicas

	def get_data_set_label(self):

		'''Get data set label corresponding to this data point.'''

		return self.data_set_label

	def get_kinematics_names(self):

		'''Get names of kinematic variables.'''

		return self.kinematic_names

	def get_kinematics_units(self):

		'''Get units of kinematic variables.'''

		return self.kinematic_units

	def get_kinematics_values(self):

		'''Get values of kinematic variables.'''

		return self.kinematic_values

	def get_kinematics_uncertainties(self):

		'''Get uncertainties of kinematic variables.'''

		return self.kinematic_uncertainties

	def get_kinematics_bins(self):

		'''Get bins of kinematic variables.'''

		return self.kinematic_bins

	def get_kinematics_replicas(self):

		'''Get replicas of kinematic variables.'''

		return self.kinematic_replicas



from gpddatabase.MarkdownDataObjectDataSet import MarkdownDataObjectDataSet as MarkdownFunctionalities

from gpddatabase.Exceptions import ExceptionNoField
from gpddatabase.Exceptions import ExceptionRequiredDiffSizes
from gpddatabase.Exceptions import ExceptionValuesLEQ
from gpddatabase.Exceptions import ExceptionValuesGEQ

from gpddatabase.DataObjectKinematics import DataObjectKinematics
from gpddatabase.DataObjectObservable import DataObjectObservable
from gpddatabase.DataObjectDataPoint import DataObjectDataPoint

class DataObjectDataSet(MarkdownFunctionalities):

	'''Class representing a single data set.'''

	def __init__(self, data, label):

		#label
		self.label = label

		#kinematics
		if 'kinematics' not in data:
			raise ExceptionNoField('kinematics')

		self.kinematics = DataObjectKinematics(data['kinematics'])

		#observables
		if 'observable' not in data:
			raise ExceptionNoField('observable')

		self.observables = DataObjectObservable(data['observable'])

		#check number of points in both sets
		if len(self.kinematics.get_values()) != len(self.observables.get_values()):
			raise ExceptionRequiredDiffSizes(self.kinematics.get_values(), self.observables.get_values())

	def get_label(self):

		'''Get label.'''

		return self.label

	def get_kinematics(self):

		'''Get object representing kinematics.'''

		return self.kinematics

	def get_observables(self):

		'''Get object representing observables.'''

		return self.observables

	def get_number_of_data_points(self):

		'''Get number of data points.'''

		return len(self.kinematics.get_values())

	def get_data_point(self, i):

		'''Get a single data point.'''

		#check index
		if i < 0:
			raise ExceptionValuesLEQ(i, -1)

		if i >= len(self.kinematics.get_values()):
			raise ExceptionValuesGEQ(i, len(self.kinematics.get_values()))

		return DataObjectDataPoint(
			self.label,

			self.kinematics.get_names(),
			self.kinematics.get_units(),
			self.kinematics.get_values()[i],
			self.kinematics.get_uncertainties()[i] if self.kinematics.get_uncertainties() else None,
			self.kinematics.get_bins()[i] if self.kinematics.get_bins() else None,
			self.kinematics.get_replicas()[i] if self.kinematics.get_replicas() else None,

			self.observables.get_names(),
			self.observables.get_units(),
			self.observables.get_values()[i],
			self.observables.get_stat_uncertainties()[i] if self.observables.get_stat_uncertainties() else None,
			self.observables.get_sys_uncertainties()[i] if self.observables.get_sys_uncertainties() else None,
			self.observables.get_sys_uncertainties_contribution_labels() if self.observables.get_sys_uncertainties_contribution_labels() else None,
			self.observables.get_sys_uncertainties_contribution()[i] if self.observables.get_sys_uncertainties_contribution() else None,
			self.observables.get_norm_uncertainties() if self.observables.get_norm_uncertainties() else None,
			self.observables.get_norm_uncertainties_contribution_labels() if self.observables.get_norm_uncertainties_contribution_labels() else None,
			self.observables.get_norm_uncertainties_contribution() if self.observables.get_norm_uncertainties_contribution() else None,
			self.observables.get_replicas()[i] if self.observables.get_replicas() else None
			)



#since object members in this class are created dynamically, disable pylint error with:
#pylint: disable=no-member
class MarkdownDataObjectUncertainty:

	''' Markdown representation of DataObjectUncertainty class.'''

	def convert_to_markdown(self):

		''' Get markdown representation.'''

		output = ''

		if self.is_asymmetric():
			output += str(self.unc_lower) + '/' + str(self.unc_upper)
		else:
			output += str(self.unc_lower)

		return output



#since object members in this class are created dynamically, disable pylint error with:
#pylint: disable=no-member
class MarkdownDataObjectReplicaSet:

	''' Markdown representation of DataObjectReplicaSet class.'''

	def convert_to_markdown(self):

		''' Get markdown representation.'''

		output = ''

		return output



#since object members in this class are created dynamically, disable pylint error with:
#pylint: disable=no-member
class MarkdownDataObjectDataPoint:

	''' Markdown representation of DataObjectDataPoint class.'''

	def convert_to_markdown(self):

		''' Get markdown representation.'''

		output = ''

		output += '| '

		for value in self.kinematic_values:
			output += str(value) + ' | '

		for i, item in enumerate(self.observables_values):
			output += str(item) + ' | '
			if self.observables_stat_uncertainties:
				output += self.observables_stat_uncertainties.get_uncertainty(i).convert_to_markdown() + ' | '
			else:
				output += "---" + ' | '

			if self.observables_sys_uncertainties:
				output += self.observables_sys_uncertainties.get_uncertainty(i).convert_to_markdown() + ' | '
			else:
				output += "---" + ' | '

		return output



from gpddatabase.MarkdownDataObjectUncertainty import MarkdownDataObjectUncertainty as MarkdownFunctionalities

from gpddatabase.Exceptions import ExceptionValuesLT
from gpddatabase.Exceptions import ExceptionWrongLength
from gpddatabase.Exceptions import ExceptionWrongType
from gpddatabase.Exceptions import ExceptionNoField
from gpddatabase.Exceptions import ExceptionAsymmetricUnc

class DataObjectUncertainty(MarkdownFunctionalities):

	'''Class representing a single uncertainty.'''

	def __init__(self, data):

		#get
		if data is None:
			raise ExceptionNoField('unc')

		if isinstance(data, (int, float)):

			if data < 0.:
				raise ExceptionValuesLT(data, 0.)

			self.unc_lower = data
			self.unc_upper = data

		elif isinstance(data, list):

			if len(data) != 2:
				raise ExceptionWrongLength('unc', 2)

			for value in data:

				if (not isinstance(value, int)) and (not isinstance(value, float)):
					raise ExceptionWrongType('unc')

				if value < 0.:
					raise ExceptionValuesLT(value, 0.)

			if len(data) == 2:
				self.unc_lower = data[0]
				self.unc_upper = data[1]

		else:

			raise ExceptionWrongType('unc')

	def get_unc_lower(self):

		'''Get lower value of uncertainty.'''

		return self.unc_lower

	def get_unc_upper(self):

		'''Get upper value of uncertainty.'''

		return self.unc_upper

	def is_asymmetric(self):

		'''Returns true is uncertainty asymmetric.'''

		return self.unc_lower != self.unc_upper

	def get_unc(self):

		'''Get value of uncertainty. If asymmetric exception will be raised.'''

		if self.is_asymmetric():
			raise ExceptionAsymmetricUnc()

		return self.unc_upper



from gpddatabase.MarkdownDataObjectCorrelation import MarkdownDataObjectCorrelation as MarkdownFunctionalities

from gpddatabase.DataObjectCorrelationMatrix import DataObjectCorrelationMatrix

from gpddatabase.Exceptions import ExceptionUnknownLabel

class DataObjectCorrelation(MarkdownFunctionalities):

	'''Class representing a set of correlation matrices.'''

	def __init__(self, data):

		#get
		self.correlation_matrices = {}

		for rawMatrix in data:

			matrix = DataObjectCorrelationMatrix(rawMatrix)
			self.correlation_matrices.update({matrix.get_label(): matrix.get_correlation_matrix()})

	def get_correlation_matrices(self):

		'''Get all correlation matrices.'''

		return self.correlation_matrices

	def get_correlation_matrix_labels(self):

		'''Get correlation matrix labels.'''

		return list(self.correlation_matrices.keys())

	def get_correlation_matrix(self, label):

		'''Get a given correlation matrix by label.'''

		if label not in self.correlation_matrices:
			raise ExceptionUnknownLabel(label)

		return self.correlation_matrices[label]



from gpddatabase.MarkdownDataObjectObservable import MarkdownDataObjectObservable as MarkdownFunctionalities

from gpddatabase.Exceptions import ExceptionNoField
from gpddatabase.Exceptions import ExceptionNotUnique
from gpddatabase.Exceptions import ExceptionRequiredDiffSizes
from gpddatabase.Exceptions import ExceptionDifferentUnitGroups
from gpddatabase.Exceptions import ExceptionWrongType
from gpddatabase.Exceptions import ExceptionWrongLength
from gpddatabase.Exceptions import ExceptionWrongLabel

import gpddatabase as db

from gpddatabase.DataObjectUncertaintySet import DataObjectUncertaintySet
from gpddatabase.DataObjectReplicaSet import DataObjectReplicaSet

class DataObjectObservable(MarkdownFunctionalities):

	'''Class representing collection of data observables.'''

	def __init__(self, data):

		#get names of variables
		self.names = []

		if 'name' not in data:
			raise ExceptionNoField('name')

		if data['name'] is None:
			raise ExceptionNoField('name')

		for name in data['name']:

			db.ExclusiveDatabase().get_observable_types().check_type(name)

			if name in self.names:
				raise ExceptionNotUnique(name)

			self.names.append(name)

		#get units
		self.units = []

		if 'unit' not in data:
			raise ExceptionNoField('unit')

		if data['unit'] is None:
			raise ExceptionNoField('unit')

		if len(data['name']) != len(data['unit']):
			raise ExceptionRequiredDiffSizes(data['name'], data['unit'])

		for unit in data['unit']:
			db.ExclusiveDatabase().get_unit_types().check_type(unit)

		for name, unit in zip(data['name'], data['unit']):

			if db.ExclusiveDatabase().get_observable_types().get_unit_group(name) != db.ExclusiveDatabase().get_unit_types().get_unit_group(unit):
				raise ExceptionDifferentUnitGroups(name, unit)

			self.units.append(unit)

		#get values
		self.values = []

		if 'value' not in data:
			raise ExceptionNoField('value')

		if data['value'] is None:
			raise ExceptionNoField('value')

		for point in data['value']:

			if len(point) != len(self.names):
				raise ExceptionRequiredDiffSizes(point, self.names)

			for value in point:
				if (not isinstance(value, int)) and (not isinstance(value, float)):
					raise ExceptionWrongType('value')

			self.values.append(point)

		#get norm unc
		if 'norm_unc' not in data:
			self.norm_uncertainties = None
		else:

			if data['norm_unc'] is None:
				raise ExceptionNoField('norm_unc')

			self.norm_uncertainties = DataObjectUncertaintySet(data['norm_unc'])

			if len(self.norm_uncertainties.get_uncertainties()) != len(self.names):
				raise ExceptionRequiredDiffSizes(self.norm_uncertainties.get_uncertainties(), self.names)

		#get norm unc contrib labels
		if 'norm_unc_contrib_label' not in data:
			self.norm_uncertainties_contrib_labels = None
		else:

			self.norm_uncertainties_contrib_labels = []

			if data['norm_unc_contrib_label'] is None:
				raise ExceptionNoField('norm_unc_contrib_label')

			labels = data['norm_unc_contrib_label']

			for label in labels:

				if not isinstance(label, str):
					raise ExceptionWrongType('norm_unc_contrib_label')

				if len(label) > 40:
					raise ExceptionWrongLength('norm_unc_contrib_label', 40)

				if not label.replace('_', '').isalnum():
					raise ExceptionWrongLabel(label)

			self.norm_uncertainties_contrib_labels = labels

		#get norm unc contrib contributions
		if 'norm_unc_contrib' not in data:
			self.norm_uncertainties_contrib = None
		else:

			self.norm_uncertainties_contrib = []

			if data['norm_unc_contrib'] is None:
				raise ExceptionNoField('norm_unc_contrib')

			for uncertainty_set_raw in data['norm_unc_contrib']:

				uncertainty_set = DataObjectUncertaintySet(uncertainty_set_raw)

				if len(uncertainty_set.get_uncertainties()) != len(self.norm_uncertainties_contrib_labels):
					raise ExceptionRequiredDiffSizes(uncertainty_set.get_uncertainties(), self.norm_uncertainties_contrib_labels)

				self.norm_uncertainties_contrib.append(uncertainty_set)

			if len(self.norm_uncertainties_contrib) != len(self.names):
				raise ExceptionRequiredDiffSizes(self.norm_uncertainties_contrib, self.names)

		#get stat unc
		if 'stat_unc' not in data:
			self.stat_uncertainties = None
		else:

			self.stat_uncertainties = []

			if data['stat_unc'] is None:
				raise ExceptionNoField('stat_unc')

			for uncertainty_set_raw in data['stat_unc']:

				uncertainty_set = DataObjectUncertaintySet(uncertainty_set_raw)

				if len(uncertainty_set.get_uncertainties()) != len(self.names):
					raise ExceptionRequiredDiffSizes(uncertainty_set.get_uncertainties(), self.names)

				self.stat_uncertainties.append(uncertainty_set)

			if len(self.stat_uncertainties) != len(self.values):
				raise ExceptionRequiredDiffSizes(self.stat_uncertainties, self.values)

		#get sys unc
		if 'sys_unc' not in data:
			self.sys_uncertainties = None
		else:

			self.sys_uncertainties = []

			if data['sys_unc'] is None:
				raise ExceptionNoField('sys_unc')

			for uncertainty_set_raw in data['sys_unc']:

				uncertainty_set = DataObjectUncertaintySet(uncertainty_set_raw)

				if len(uncertainty_set.get_uncertainties()) != len(self.names):
					raise ExceptionRequiredDiffSizes(uncertainty_set.get_uncertainties(), self.names)

				self.sys_uncertainties.append(uncertainty_set)

			if len(self.sys_uncertainties) != len(self.values):
				raise ExceptionRequiredDiffSizes(self.sys_uncertainties, self.values)

		#get sys unc contrib labels
		if 'sys_unc_contrib_label' not in data:
			self.sys_uncertainties_contrib_labels = None
		else:

			self.sys_uncertainties_contrib_labels = []

			if data['sys_unc_contrib_label'] is None:
				raise ExceptionNoField('sys_unc_contrib_label')

			labels = data['sys_unc_contrib_label']

			for label in labels:

				if not isinstance(label, str):
					raise ExceptionWrongType('sys_unc_contrib_label')

				if len(label) > 40:
					raise ExceptionWrongLength('sys_unc_contrib_label', 40)

				if not label.replace('_', '').isalnum():
					raise ExceptionWrongLabel(label)

			self.sys_uncertainties_contrib_labels = labels

		#get sys unc contrib
		if 'sys_unc_contrib' not in data:
			self.sys_uncertainties_contrib = None
		else:

			self.sys_uncertainties_contrib = []

			if data['sys_unc_contrib'] is None:
				raise ExceptionNoField('sys_unc_contrib')

			for uncertainty_set_set_raw in data['sys_unc_contrib']:

				uncertainty_set_set = []

				if isinstance(uncertainty_set_set_raw[0], str):

					uncertainty_set = DataObjectUncertaintySet(uncertainty_set_set_raw)

					if len(uncertainty_set.get_uncertainties()) != (len(self.sys_uncertainties_contrib_labels) * len(self.names)):
						raise ExceptionRequiredDiffSizes(uncertainty_set.get_uncertainties(), (len(self.sys_uncertainties_contrib_labels) * len(self.names)))

					uncertainty_set_set.append(uncertainty_set)

				else:

					for uncertainty_set_raw in uncertainty_set_set_raw:

						uncertainty_set = DataObjectUncertaintySet(uncertainty_set_raw)

						if len(uncertainty_set.get_uncertainties()) != len(self.sys_uncertainties_contrib_labels):
							raise ExceptionRequiredDiffSizes(uncertainty_set.get_uncertainties(), self.sys_uncertainties_contrib_labels)

						uncertainty_set_set.append(uncertainty_set)

					if len(uncertainty_set_set) != len(self.names):
						raise ExceptionRequiredDiffSizes(uncertainty_set_set, self.names)

				self.sys_uncertainties_contrib.append(uncertainty_set_set)

			if len(self.sys_uncertainties_contrib) != len(self.values):
				raise ExceptionRequiredDiffSizes(self.sys_uncertainties_contrib, self.values)

		#get replicas
		if 'replica' not in data:
			self.replicas = None
		else:

			self.replicas = []

			if data['replica'] is None:
				raise ExceptionNoField('replica')

			for replicas in data['replica']:

				this_replicas = []

				if len(replicas) != len(self.names):
					raise ExceptionRequiredDiffSizes(replicas, self.names)

				for replica in replicas:
					this_replicas.append(DataObjectReplicaSet(replica))

				self.replicas.append(this_replicas)

			if len(self.replicas) != len(self.values):
				raise ExceptionRequiredDiffSizes(self.replicas, self.values)

	def get_names(self):

		'''Get names of variables.'''

		return self.names

	def get_units(self):

		'''Get names and units.'''

		return self.units

	def get_values(self):

		'''Get values.'''

		return self.values

	def get_stat_uncertainties(self):

		'''Get statistical uncertainties.'''

		return self.stat_uncertainties

	def get_sys_uncertainties(self):

		'''Get systematic uncertainties.'''

		return self.sys_uncertainties

	def get_sys_uncertainties_contribution_labels(self):

		'''Get labels of contributions to systematic uncertainties.'''

		return self.sys_uncertainties_contrib_labels

	def get_sys_uncertainties_contribution(self):

		'''Get contributions to systematic uncertainties.'''

		return self.sys_uncertainties_contrib

	def get_norm_uncertainties(self):

		'''Get normalisation uncertainties.'''

		return self.norm_uncertainties

	def get_norm_uncertainties_contribution_labels(self):

		'''Get labels of contributions to normalisation uncertainties.'''

		return self.norm_uncertainties_contrib_labels

	def get_norm_uncertainties_contribution(self):

		'''Get contributions to normalisation uncertainties.'''

		return self.norm_uncertainties_contrib

	def get_replicas(self):

		'''Get replicas.'''

		return self.replicas



from gpddatabase.MarkdownDataObjectReplicaSet import MarkdownDataObjectReplicaSet as MarkdownFunctionalities

from gpddatabase.Exceptions import ExceptionNoField
from gpddatabase.Exceptions import ExceptionWrongType

class DataObjectReplicaSet(MarkdownFunctionalities):

	'''Class representing replica set.'''

	def __init__(self, data):

		#values
		self.values = []

		if data is None:
			raise ExceptionNoField('replica')

		for value in data:

			if (not isinstance(value, int)) and (not isinstance(value, float)):
				raise ExceptionWrongType('replica')

			self.values.append(value)

	def get_values(self):

		'''Get replica values.'''

		return self.values



from particle import Particle
from particle import ParticleNotFound

from gpddatabase.Exceptions import ExceptionUnknownType

class ParticleTypes:

	'''Class defining particle types. Uses python 'particle' library.'''

	def check_type(self, value):

		'''Check if type exist. If not, raise exception.'''

		self.get_particle(value)

	def get_description(self, value):

		'''Get description of a given type.'''

		return self.get_particle(value).name

	def get_particle(self, value):

		'''Get 'Particle' object (see 'particle' library) for a given type.'''

		for particle in Particle.findall(value):
			if particle.name == value:
				return particle

		raise ExceptionUnknownType(value) from ParticleNotFound



#since object members in this class are created dynamically, disable pylint error with:
#pylint: disable=no-member
class MarkdownDataObjectCorrelation:

	''' Markdown representation of MarkdownDataObjectCorrelation class.'''

	def convert_to_markdown(self):

		''' Get markdown representation.'''

		output = ''

		output += '## Definition of correlation matrices:' + '\n'

		output += '\n'

		for label in self.correlation_matrices:
			output += '* ' +  label + ': ' + '\n'
			output += str(self.correlation_matrices[label]) + '\n'

		return output



class ExceptionNoDataInFile(Exception):

	'''Exception signaling no data in given file.'''

	def __init__(self, value):
		self.value = value

	def __str__(self):
		return "No data in file " + self.value

class ExceptionNoField(Exception):

	'''Exception signaling no filed of a given name in yaml file.'''

	def __init__(self, value):
		self.value = value

	def __str__(self):
		return "Field " + self.value + " is missing"

class ExceptionNotUnique(Exception):

	'''Exception signaling value being not unique.'''

	def __init__(self, value):
		self.value = value

	def __str__(self):
		return "Value " + self.value + " is not unique"

class ExceptionUnknownType(Exception):

	'''Exception signaling unknown type.'''

	def __init__(self, value):
		self.value = value

	def __str__(self):
		return "Unknown type " + str(self.value)

class ExceptionWrongType(Exception):

	'''Exception signaling wrong type.'''

	def __init__(self, value):
		self.value = value

	def __str__(self):
		return "Wrong type of field " + str(self.value)

class ExceptionWrongUUID(Exception):

	'''Exception signaling wrong UUID.'''

	def __str__(self):
		return "UUID must have exactly 8 alphanumeric characters"

class ExceptionWrongLabel(Exception):

	'''Exception signaling wrong label.'''

	def __str__(self):
		return "Label must only have alphanumeric characters and/or '_' character"


class ExceptionNoDirectory(Exception):

	'''Exception signaling field not set.'''

	def __init__(self, value):
		self.value = value

	def __str__(self):
		return "Directory " + str(self.value) + " does not exist"

class ExceptionNotUniqueUUID(Exception):

	'''Exception signaling not unique UUID.'''

	def __init__(self, value):
		self.value = value

	def __str__(self):
		return "UUID of " + str(self.value) + " is not unique"

class ExceptionUnknownUUID(Exception):

	'''Exception signaling unknown UUID.'''

	def __init__(self, value):
		self.value = value

	def __str__(self):
		return "Unknown UUID " + str(self.value)

class ExceptionStringToDate(Exception):

	'''Exception signaling bad conversion from string to date.'''

	def __init__(self, value):
		self.value = value

	def __str__(self):
		return "Unable to convert " + str(self.value) + " into date"

class ExceptionWrongLength(Exception):

	'''Exception signaling wrong length of string.'''

	def __init__(self, value1, value2):
		self.value1 = value1
		self.value2 = value2

	def __str__(self):
		return "Field " + str(self.value1) + " exceeds allowed length " + str(self.value2)

class ExceptionNotWritable(Exception):

	'''Exception signaling not writable directory.'''

	def __init__(self, value):
		self.value = value

	def __str__(self):
		return "Directory " + str(self.value) + " is not writable"

class ExceptionNoRequirement(Exception):

	'''Exception signaling missing field in 'conditions' section.'''

	def __init__(self, value):
		self.value = value

	def __str__(self):
		return "Field " + str(self.value) + " required by this data type in missing in 'conditions' section"

class ExceptionRequiredDiffSizes(Exception):

	'''Exception signaling different sizes of two objects.'''

	def __init__(self, valueA, valueB):
		self.valueA = valueA
		self.valueB = valueB

	def __str__(self):
		return "Different sizes of " + str(self.valueA) + " and " + str(self.valueB)

class ExceptionUnknownLabel(Exception):

	'''Exception signaling unknown label.'''

	def __init__(self, value):
		self.value = value

	def __str__(self):
		return "Unknown label " + str(self.value)

class ExceptionWrongCorrelationMatrix(Exception):

	'''Exception signaling wrong correlation matrix.'''

	def __init__(self, value):
		self.value = value

	def __str__(self):
		return "Wrong definition of correlation matrix " + str(self.value)

class ExceptionDifferentUnitGroups(Exception):

	'''Exception signaling different data groups.'''

	def __init__(self, valueA, valueB):
		self.valueA = valueA
		self.valueB = valueB

	def __str__(self):
		return "Variable " + str(self.valueA) + " does not belong to variable group of " + str(self.valueB)

class ExceptionValuesGEQ(Exception):

	'''Exception signaling valueA >= valueB'''

	def __init__(self, valueA, valueB):
		self.valueA = valueA
		self.valueB = valueB

	def __str__(self):
		return "Value " + str(self.valueA) + " grater or equal value " + str(self.valueB)

class ExceptionValuesLEQ(Exception):

	'''Exception signaling valueA <= valueB'''

	def __init__(self, valueA, valueB):
		self.valueA = valueA
		self.valueB = valueB

	def __str__(self):
		return "Value " + str(self.valueA) + " smaller or equal value " + str(self.valueB)

class ExceptionValuesGT(Exception):

	'''Exception signaling valueA > valueB'''

	def __init__(self, valueA, valueB):
		self.valueA = valueA
		self.valueB = valueB

	def __str__(self):
		return "Value " + str(self.valueA) + " grater than value " + str(self.valueB)

class ExceptionValuesLT(Exception):

	'''Exception signaling valueA < valueB'''

	def __init__(self, valueA, valueB):
		self.valueA = valueA
		self.valueB = valueB

	def __str__(self):
		return "Value " + str(self.valueA) + " smaller than value " + str(self.valueB)

class ExceptionAsymmetricUnc(Exception):

	'''Exception signaling asymmetric uncertainty.'''

	def __str__(self):
		return "This uncertainty is asymmetric, use different function"



#since object members in this class are created dynamically, disable pylint error with:
#pylint: disable=no-member
class MarkdownDataObjectDataSet:

	''' Markdown representation of DataObjectDataSet class.'''

	def convert_to_markdown(self):

		''' Get markdown representation.'''

		output = ''
		additional_info = ''
		table_line = ''
		unit_line = ''
		scale_line = ''

		output += '### Data set: ' + self.label + '\n'

		output += '\n'

		# rather than running
		# output += self.kinematics.convert_to_markdown()
		# output += self.observables.convert_to_markdown()
		# we loop over data points

		if self.kinematics.get_uncertainties():
			additional_info += '* uncertainties of kinematic values' + '\n'
		if self.kinematics.get_replicas():
			additional_info += '* replica values for kinematics' + '\n'
		if self.kinematics.get_bins():
			additional_info += '* boundaries of kinematic bins' + '\n'
		if self.observables.get_norm_uncertainties_contribution():
			additional_info += '* contributions to normalisation uncertainties' + '\n'
		if self.observables.get_sys_uncertainties_contribution():
			additional_info += '* contributions to systematic uncertainties' + '\n'
		if self.observables.get_replicas():
			additional_info += '* replica values for observables' + '\n'
		if self.observables.get_norm_uncertainties():
			additional_info += '* correlation between normalisation uncertainties' + '\n'

		if self.observables.get_stat_uncertainties():

			hasStatCorr = False

			for uncertainty_set in self.observables.get_stat_uncertainties():
				if uncertainty_set.get_correlation_matrix():
					hasStatCorr = True
					break

			if hasStatCorr:
				additional_info += '* correlation between statistical uncertainties' + '\n'

		if self.observables.get_sys_uncertainties():

			hasStatCorr = False

			for uncertainty_set in self.observables.get_sys_uncertainties():
				if uncertainty_set.get_correlation_matrix():
					hasStatCorr = True
					break

			if hasStatCorr:
				additional_info += '* correlation between systematic uncertainties' + '\n'

		if additional_info != '':
			output += 'Additional information availible:' + '\n'
			output += additional_info


		output += '\n'

		output += 'Table of values:' + '\n'

		output += '\n'

		output += '| | '
		table_line += '| --- | '
		unit_line += '| unit: | '
		if self.observables.get_norm_uncertainties():
			scale_line += '| norm unc.: | '

		for name, unit in zip(self.kinematics.get_names(), self.kinematics.get_units()):
			output += name + ' | '
			table_line += ':---:' + ' | '
			unit_line += unit + ' | '
			if self.observables.get_norm_uncertainties():
				scale_line += ' ' + ' | '

		for name, unit in zip(self.observables.get_names(), self.observables.get_units()):
			output += name + ' | stat | sys | '
			table_line += ':---:' + ' | :---: | :---: | '
			unit_line += unit + ' | ' + unit + ' | ' + unit + ' | '

		if self.observables.get_norm_uncertainties():
			for i in range(0, self.observables.get_norm_uncertainties().get_number_of_uncertainties()):
				scale_line += self.observables.get_norm_uncertainties().get_uncertainty(i).convert_to_markdown() + ' |  |  | '

		output += '\n' + table_line
		output += '\n' + unit_line
		if self.observables.get_norm_uncertainties():
			output += '\n' + scale_line
		output += '\n'

		for i in range(0, self.get_number_of_data_points()):
			output += '| ' + str(i) + self.get_data_point(i).convert_to_markdown() + '\n'

		output += '\n'

		return output



from munch import Munch as munch

from gpddatabase.MarkdownDataObjectGeneralInfo import MarkdownDataObjectGeneralInfo as MarkdownFunctionalities

from gpddatabase.Exceptions import ExceptionNoField
from gpddatabase.Exceptions import ExceptionStringToDate
from gpddatabase.Exceptions import ExceptionWrongType
from gpddatabase.Exceptions import ExceptionWrongLength
from gpddatabase.Exceptions import ExceptionNoRequirement

import gpddatabase as db

from gpddatabase.ParticleTypes import ParticleTypes

class DataObjectGeneralInfo(MarkdownFunctionalities):

	'''Class storing general information.'''

	def __init__(self, data):

		#date
		if 'date' not in data:
			raise ExceptionNoField('date')

		if data['date'] is None:
			raise ExceptionNoField('date')

		try:
			data['date'].fromisoformat('1900-01-01')
		except AttributeError as err:
			raise ExceptionStringToDate(data['date']) from err

		self.date = data['date']

		#data type
		if 'data_type' not in data:
			raise ExceptionNoField('data_type')

		if data['data_type'] is None:
			raise ExceptionNoField('data_type')

		db.ExclusiveDatabase().get_data_types().check_type(data['data_type'])

		self.data_type = data['data_type']

		#pseudodata
		if 'pseudodata' not in data:
			self.pseudodata = False
		else:

			if data['pseudodata'] is None:
				raise ExceptionNoField('data_type')

			if not isinstance(data['pseudodata'], bool):
				raise ExceptionWrongType('pseudodata')

			self.pseudodata = data['pseudodata']

		#collaboration
		if 'collaboration' not in data:
			raise ExceptionNoField('collaboration')

		if data['collaboration'] is None:
			raise ExceptionNoField('collaboration')

		if not isinstance(data['collaboration'], str):
			raise ExceptionWrongType('collaboration')

		if len(data['collaboration']) > 40:
			raise ExceptionWrongLength('collaboration', 40)

		self.collaboration = data['collaboration']

		#reference
		if 'reference' not in data:
			self.reference = None
		else:

			if data['reference'] is None:
				raise ExceptionNoField('reference')

			if not isinstance(data['reference'], str):
				raise ExceptionWrongType('reference')

			if len(data['reference']) > 255:
				raise ExceptionWrongLength('reference', 255)

			self.reference = data['reference']

		#conditions
		if 'conditions' not in data:
			raise ExceptionNoField('conditions')

		if data['conditions'] is None:
			raise ExceptionNoField('conditions')

		self.conditions = munch.toDict(data['conditions'])

		for requirement in db.ExclusiveDatabase().get_data_types().get_required_name(self.data_type):
			if requirement not in self.conditions:

				if requirement == 'hadron_beam_energy':
					self.conditions['hadron_beam_energy'] = 0.001 * ParticleTypes().get_particle(self.conditions['hadron_beam_type']).mass # MeV -> GeV
				else:
					raise ExceptionNoRequirement(requirement)

			else:
				db.ExclusiveDatabase().get_required_types().check_value(
					db.ExclusiveDatabase().get_data_types().get_required_type_by_name(self.data_type, requirement),
					self.conditions[requirement]
				)

		for requirement in db.ExclusiveDatabase().get_data_types().get_optional_name(self.data_type):
			if requirement not in self.conditions:
				self.conditions[requirement] = db.ExclusiveDatabase().get_data_types().get_optional_default_by_name(self.data_type, requirement)

			else:
				db.ExclusiveDatabase().get_required_types().check_value(
					db.ExclusiveDatabase().get_data_types().get_optional_type_by_name(self.data_type, requirement),
					self.conditions[requirement]
				)

		#comment
		if 'comment' not in data:
			self.comment = None
		else:

			if data['comment'] is None:
				raise ExceptionNoField('comment')

			if not isinstance(data['comment'], str):
				raise ExceptionWrongType('comment')

			if len(data['comment']) > 255:
				raise ExceptionWrongLength('comment', 255)

			self.comment = data['comment']

	def get_date(self):

		'''Get date.'''

		return self.date

	def get_data_type(self):

		'''Get data type.'''

		return self.data_type

	def get_pseudodata(self):

		'''Check if pseudodata.'''

		return self.pseudodata

	def get_collaboration(self):

		'''Get collaboration name.'''

		return self.collaboration

	def get_reference(self):

		'''Get reference.'''

		return self.reference

	def get_conditions(self):

		'''Get conditions.'''

		return self.conditions

	def get_comment(self):

		'''Get comment.'''

		return self.comment



from gpddatabase.MarkdownDataObjectUUID import MarkdownDataObjectUUID as MarkdownFunctionalities

from gpddatabase.Exceptions import ExceptionNoField
from gpddatabase.Exceptions import ExceptionWrongType
from gpddatabase.Exceptions import ExceptionWrongUUID

class DataObjectUUID(MarkdownFunctionalities):

	'''Class representing UUID.'''

	def __init__(self, uuid):

		if uuid is None:
			raise ExceptionNoField('uuid')

		if not isinstance(uuid, str):
			raise ExceptionWrongType('uuid')

		if len(uuid) != 8 or (not uuid.isalnum()):
			raise ExceptionWrongUUID()

		self.uuid = uuid

	def get_uuid(self):

		'''Get UUID value.'''

		return self.uuid
